from threading import Event
from queue import PriorityQueue, Empty
from time import sleep
from datetime import datetime, timezone

from yaspin import Spinner


def print_logs_from_queue(log_queue: PriorityQueue, stop_event: Event, spinner: Spinner):
    last_log_timestamp = None
    last_log_printed_at = None
    is_first_log = True
    while True:
        try:
            timestamp, log_message = log_queue.get_nowait()

            if is_first_log:
                sleep(6)  # remote_parallel_map.JOB_STATUS_POLL_RATE_SEC
                is_first_log = False

            current_epoch = datetime.now(timezone.utc).timestamp()

            last_log_printed_delta = current_epoch - (last_log_printed_at or current_epoch)
            last_log_occurred_delta = timestamp - (last_log_timestamp or timestamp)
            sleep_time = max(last_log_occurred_delta - last_log_printed_delta, 0)

            # if last_log_occurred_delta - last_log_printed_delta < 0:
            #     lateness = (last_log_occurred_delta - last_log_printed_delta) * -1
            #     print(f"LATE!! Log should have printed {lateness} seconds ago!")

            sleep(sleep_time)
            spinner.write(log_message)
            last_log_printed_at = datetime.now(timezone.utc).timestamp()
            last_log_timestamp = timestamp
        except Empty:
            if stop_event.is_set():
                return
            sleep(0.5)



import os
import sys
import requests
import warnings
import pickle
from six import reraise
from threading import Thread, Event
from typing import Callable, Optional
from time import sleep, time
from queue import PriorityQueue

import cloudpickle
from yaspin import yaspin, Spinner
from tblib import pickling_support

from burla import _BURLA_SERVICE_URL
from burla._logstream import print_logs_from_queue
from burla._env_inspection import get_pip_packages, get_function_dependencies
from burla._auth import auth_headers_from_local_config, AuthException, login_required
from burla._helpers import (
    nopath_warning,
    JobTimeoutError,
    InstallError,
    ServerError,
    StatusMessage,
    upload_inputs,
    download_outputs,
)

pickling_support.install()
warnings.formatwarning = nopath_warning


MAX_CPUS = 2000  # please ask before you increase this <3
MAX_GPUS = 300  # please ask before you increase this <3

JOB_STATUS_POLL_RATE_SEC = 6  # how often to check job status
TIMEOUT_MIN = 60 * 12  # max time a Burla job can run for
IN_COLAB = os.getenv("COLAB_RELEASE_TAG") is not None

BYTES_HEADER = {"Content-Type": "application/octet-stream"}


def _get_job_info_since(epoch: int, job_id: str, headers: dict):
    """Raises: ServerError, InstallError, HTTPError, or re-raised error from UDF"""
    response = requests.get(f"{_BURLA_SERVICE_URL}/v1/jobs/{job_id}/{epoch}", headers=headers)
    if response.status_code == 401:
        AuthException()
    elif response.status_code >= 500:
        raise ServerError()
    elif response.status_code >= 300:
        response.raise_for_status()

    job = response.json()
    if job.get("udf_error"):
        error = pickle.loads(bytes.fromhex(job.get("udf_error")))
        reraise(*error)
    elif job.get("install_error"):
        raise InstallError(job["install_error"])

    if job.get("output_urls"):
        return job["udf_started"], job["logs"], job["output_urls"]
    else:
        return job["udf_started"], job["logs"], None


def _start_job(
    function_: Callable,
    inputs: list,
    auth_headers: dict,
    verbose: bool,
    spinner: Spinner,
    func_cpu: int,
    func_ram: int,
    func_gpu: int,
    parallelism: int,
    image: Optional[str] = None,
) -> str:
    if verbose:
        spinner.text = StatusMessage.preparing()
        spinner.start()

    installed_packages = list(get_pip_packages())
    imported_modules = list(get_function_dependencies(function_))
    required_packages = [pkg for pkg in installed_packages if pkg["name"] in imported_modules]

    payload = {
        "n_inputs": len(inputs),
        "image": image,
        "func_cpu": func_cpu,
        "func_ram": func_ram,
        "func_gpu": func_gpu,
        "parallelism": parallelism,
        "python_version": f"3.8" if sys.version_info.minor < 8 else f"3.{sys.version_info.minor}",
        "packages": None if image else required_packages,
    }
    response = requests.post(f"{_BURLA_SERVICE_URL}/v1/jobs/", json=payload, headers=auth_headers)
    if response.status_code == 401:
        raise AuthException()
    else:
        response.raise_for_status()
        response_json = response.json()
        job_id = response_json["job_id"]

    # download list of urls to upload to, these urls are big and sometimes there are a lot of them.
    response = requests.get(response_json["input_urls_url"], headers=BYTES_HEADER)
    response.raise_for_status()
    input_urls = pickle.loads(response.content)

    spinner.text = StatusMessage.uploading_inputs
    upload_inputs(input_urls, inputs)

    spinner.text = StatusMessage.uploading_function
    function_pkl = cloudpickle.dumps(function_)
    requests.put(response_json["function_url"], headers=BYTES_HEADER, data=function_pkl)

    spinner.text = StatusMessage.preparing()
    response = requests.post(f"{_BURLA_SERVICE_URL}/v1/jobs/{job_id}", headers=auth_headers)
    response.raise_for_status()
    return job_id


def _watch_job(job_id: str, job_started_time: int, headers: dict, verbose: bool, spinner: Spinner):
    last_epoch = job_started_time
    epoch = last_epoch
    output_urls = None
    job_timed_out = False

    # Start printing logs generated by this job using a separate thread.
    print_queue = PriorityQueue()
    stop_event = Event()
    args = (print_queue, stop_event, spinner)
    log_thread = Thread(target=print_logs_from_queue, args=args, daemon=True)
    log_thread.start()
    while (not output_urls) and (not job_timed_out):
        sleep(JOB_STATUS_POLL_RATE_SEC)
        udf_started, logs, output_urls = _get_job_info_since(last_epoch, job_id, headers)

        for epoch, log_message in logs:
            print_queue.put((epoch, log_message))

        if verbose and udf_started:
            spinner.text = StatusMessage.running()

        last_epoch = epoch
        job_timed_out = (time() - job_started_time) > (TIMEOUT_MIN * 60)

    stop_event.set()
    log_thread.join()

    if job_timed_out:
        raise JobTimeoutError(job_id=job_id, timeout=TIMEOUT_MIN)
    return output_urls


@login_required
def remote_parallel_map(
    function_: Callable,
    inputs: list,
    verbose: bool = True,
    image: Optional[str] = None,
    func_cpu: int = 1,
    func_ram: int = 1,
    func_gpu: int = 0,
    parallelism: int = -1,
):
    if (func_cpu > 96) or (func_cpu < 1):
        raise ValueError("CPU per function call must be one of [1.. 96]")
    if (func_ram > 624) or (func_ram < 1):
        raise ValueError("RAM per function call must be one of [1.. 624]")
    if (func_gpu > 4) or (func_gpu < 0):
        raise ValueError("GPU per function call must be one of [0.. 4]")

    n_inputs = len(inputs)
    parallelism = n_inputs if parallelism == -1 else parallelism
    requested_cpu = parallelism * func_cpu
    requested_gpu = parallelism * func_gpu

    if (requested_cpu > MAX_CPUS) and not (requested_gpu > MAX_GPUS):
        parallelism = MAX_CPUS // func_cpu
        warnings.warn(
            f"Limiting parallelism to {parallelism} to stay under current limit of {MAX_CPUS} CPUs."
        )
    if requested_gpu > MAX_GPUS:
        parallelism = MAX_GPUS // func_gpu
        warnings.warn(
            f"Limiting parallelism to {parallelism} to stay under current limit of {MAX_GPUS} GPUs."
        )

    max_concurrency = 10000
    input_batches = [inputs[i : i + max_concurrency] for i in range(0, n_inputs, max_concurrency)]
    if len(input_batches) > 1:
        msg = f"Because a maximum of {max_concurrency} inputs can be processed at a time, "
        msg += f"these inputs will be processed in {len(input_batches)} separate batches. "
        msg += "Consider using a lower number of larger inputs, this may reduce runtime."
        warnings.warn(msg)

    spinner = yaspin()
    StatusMessage.function_name = function_.__name__
    StatusMessage.n_inputs = len(inputs) if len(inputs) < max_concurrency else max_concurrency
    StatusMessage.total_cpus = parallelism * func_cpu
    StatusMessage.total_gpus = parallelism * func_gpu

    auth_headers = auth_headers_from_local_config()

    return_values = []
    for input_batch in input_batches:
        try:
            start_time = int(time())
            job_id = _start_job(
                function_=function_,
                inputs=input_batch,
                auth_headers=auth_headers,
                verbose=verbose,
                spinner=spinner,
                func_cpu=func_cpu,
                func_ram=func_ram,
                func_gpu=func_gpu,
                parallelism=parallelism,
                image=image,
            )
            output_urls = _watch_job(job_id, start_time, auth_headers, verbose, spinner)

            spinner.text = StatusMessage.downloading
            for output_pkl in download_outputs(output_urls):
                output = cloudpickle.loads(output_pkl)
                if output["pickled_with_cloudpickle"]:
                    return_values.append(cloudpickle.loads(output["return_value_pkl"]))
                elif output["return_type"] == "<class 'spacy.tokens.doc.Doc'>":
                    if (not globals().get("Doc")) and (not globals().get("Vocab")):
                        from spacy.tokens.doc import Doc
                        from spacy.vocab import Vocab
                    return_values.append(Doc(Vocab()).from_bytes(output["return_value_pkl"]))

        except Exception as e:
            spinner.stop()
            raise e

    if verbose:
        spinner.text = "Done!"
        spinner.ok("✔")

    all_return_values_are_none = all(value is None for value in return_values)
    if not all_return_values_are_none:
        return return_values



import os
import json
import webbrowser
import requests
from time import sleep
from uuid import uuid4
from pathlib import Path
from typing import Tuple

from appdirs import user_config_dir
from IPython.core.display import Javascript
from IPython.display import display, clear_output

from burla import _BURLA_SERVICE_URL

AUTH_TIMEOUT_SECONDS = 180
CONFIG_PATH = Path(user_config_dir(appname="burla", appauthor="burla")) / Path("credentials.json")
IN_COLAB = os.getenv("COLAB_RELEASE_TAG") is not None


class AuthTimeoutException(Exception):
    def __init__(self):
        super().__init__("Timed out waiting for authentication flow to complete.")


class AuthException(Exception):
    def __init__(self):
        super().__init__(
            "You are not logged in! Please run `burla login` to create an account or login."
        )


def login_credentials_missing() -> bool:
    return not CONFIG_PATH.exists()


def auth_headers_from_local_config() -> Tuple[str, str]:
    if login_credentials_missing():
        raise AuthException()
    else:
        auth_info = json.loads(CONFIG_PATH.read_text())
        return {"email": auth_info["email"], "Authorization": f"Bearer {auth_info['api_key']}"}


def _get_auth_creds(client_id, attempt=0):
    if attempt == AUTH_TIMEOUT_SECONDS / 2:
        raise AuthTimeoutException()

    sleep(2)
    response = requests.get(f"{_BURLA_SERVICE_URL}/v1/keys/{client_id}")

    if response.status_code == 404:
        return _get_auth_creds(client_id, attempt=attempt + 1)
    else:
        response.raise_for_status()
        return response.json()["api_key"], response.json()["email"]


def login_cmd():
    if IN_COLAB:
        raise SystemExit(
            (
                "\nUnable to login using this command from inside a Google Colab notebook!\n"
                "To login simply call `burla.remote_parallel_map`, or `burla.login`, eg:\n"
                "```\n"
                "from burla import login\n"
                "login()\n"
                "```"
            )
        )
    else:
        login()


def login():
    client_id = uuid4().hex
    login_url = f"{_BURLA_SERVICE_URL}/v1/login/{client_id}"

    if IN_COLAB:
        display(Javascript(f'window.open("{login_url}");'))
        sleep(1)  # give js a second to run before removing it
        clear_output()  # prevents js from re-running automatically when notebook opened
    else:
        webbrowser.open(login_url)

    print(f"Your browser has been opened to visit:\n\n    {login_url}\n")
    api_key, email = _get_auth_creds(client_id)

    message = f"Thank you for registering with Burla! You are now logged in as [{email}].\n"
    message += "Please email jake@burla.dev with any questions!\n"
    print(message)

    if not CONFIG_PATH.exists():
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        CONFIG_PATH.touch()
    CONFIG_PATH.write_text(json.dumps({"api_key": api_key, "email": email}))


def login_required(func):
    def wrapped(*args, **kwargs):
        if login_credentials_missing() and IN_COLAB:
            login()
        elif login_credentials_missing():
            raise AuthException()
        return func(*args, **kwargs)

    return wrapped


@login_required
def current_user():
    return json.loads(CONFIG_PATH.read_text())["email"]



import requests
import json
from typing import Optional
from pathlib import Path

from yaspin import yaspin
from appdirs import user_config_dir

from burla import _BURLA_SERVICE_URL
from burla._auth import login_required, auth_headers_from_local_config, current_user

BYTES_HEADER = {"Content-Type": "application/octet-stream"}
CONFIG_PATH = Path(user_config_dir(appname="burla", appauthor="burla"))
CURRENT_DIR_PATH = CONFIG_PATH / Path("current_dir.json")


def _get_current_dir():
    CONFIG_PATH.mkdir(parents=True, exist_ok=True)
    if not CURRENT_DIR_PATH.exists():
        CURRENT_DIR_PATH.touch()
        CURRENT_DIR_PATH.write_text(json.dumps({"current_dir": "/"}))
    return json.loads(CURRENT_DIR_PATH.read_text())["current_dir"]


@login_required
def rm(recurse: bool = False):
    pass


@login_required
def cd(folder: str):
    current_dir = _get_current_dir()
    current_dir = current_dir[:-1] if current_dir.endswith("/") else current_dir

    if folder == ".":
        print(f"{current_user()} {current_dir} % ")
        return
    elif folder == ".." and current_dir == "/":
        return
    elif folder == "..":
        target_dir = "/".join(current_dir.split("/")[:-1])
        target_dir = "/" if target_dir == "" else target_dir
    elif folder.startswith("/"):
        target_dir = folder
    else:
        target_dir = current_dir + f"/{folder}"

    url = f"{_BURLA_SERVICE_URL}/v1/bcs/object_info"
    headers = auth_headers_from_local_config()
    response = requests.post(url, json={"remote_path": target_dir}, headers=headers)
    response.raise_for_status()
    remote_object_info = response.json()
    if remote_object_info.get("type") == "folder":
        CURRENT_DIR_PATH.write_text(json.dumps({"current_dir": target_dir}))
        print(f"{current_user()} {target_dir} % ")
    elif remote_object_info.get("type") != "folder":
        print(f"burla nas cd: no such file or directory: {target_dir}")


@login_required
def ls():
    current_dir = _get_current_dir()
    current_dir = "/" if current_dir == "" else current_dir
    print(f"{current_user()} {current_dir} % ")

    url = f"{_BURLA_SERVICE_URL}/v1/bcs/object_info"
    headers = auth_headers_from_local_config()
    response = requests.post(url, json={"remote_path": current_dir}, headers=headers)
    response.raise_for_status()
    remote_object_info = response.json()
    if not remote_object_info.get("type") == "folder":
        raise Exception("Current directory is not a folder??")

    print("\t".join(remote_object_info["contents_relative_paths"]))


@login_required
def upload(local_path: str, remote_folder: Optional[str] = None, recurse: bool = False):
    local_path = Path(local_path)
    if remote_folder is None:
        remote_folder = Path(_get_current_dir())
    elif remote_folder.startswith("/"):
        remote_folder = Path(remote_folder)
    else:
        remote_folder = Path(_get_current_dir()) / Path(remote_folder)

    if not local_path.exists():
        print(f"Path {local_path} does not exist.")
        return
    elif local_path.is_dir() and (not recurse):
        msg = f"Path {local_path} is a directory, "
        msg += "please use `--recurse` to recursively upload all files in a directory."
        print(msg)
        return
    elif local_path.is_file() and recurse:
        print(f"`--recurse` option was used but path {local_path} is not a directory.")
        return
    elif not (local_path.is_dir() or local_path.is_file()):
        msg = f"Cannot upload {local_path}, is not file or folder (may be symlink or block device)"
        print(msg)
        return

    url = f"{_BURLA_SERVICE_URL}/v1/bcs/object_info"
    headers = auth_headers_from_local_config()
    response = requests.post(url, json={"remote_path": str(remote_folder)}, headers=headers)
    response.raise_for_status()
    remote_object_info = response.json()
    if remote_object_info.get("type") == "file":
        print(f"Destination must be a folder but {remote_folder} is a file.")
        return

    spinner = yaspin()
    spinner.text = f"Uploading {local_path} ..."
    spinner.start()

    try:
        if local_path.is_dir():
            filepaths = []
            local_filepaths = [subpath for subpath in local_path.rglob("*") if subpath.is_file()]
            for local_filepath in local_filepaths:
                remote_filepath = remote_folder / local_filepath
                fp = dict(remote_filepath=str(remote_filepath), local_filepath=str(local_filepath))
                filepaths.append(fp)
        elif local_path.is_file():
            remote_filepath = remote_folder / local_path.name
            filepaths = [dict(remote_filepath=str(remote_filepath), local_filepath=str(local_path))]

        headers = auth_headers_from_local_config()
        url = f"{_BURLA_SERVICE_URL}/v1/bcs/upload_urls"
        payload = {"remote_filepaths": [fp["remote_filepath"] for fp in filepaths]}
        response = requests.post(url, json=payload, headers=headers)
        response.raise_for_status()

        def read_in_chunks(file, chunk_size=8192):
            while True:
                data = file.read(chunk_size)
                if not data:
                    break
                yield data

        for remote_filepath, url in response.json()["paths_to_urls"]:
            filepath = [fp for fp in filepaths if fp["remote_filepath"] == remote_filepath][0]
            spinner.text = f"Uploading {filepath['local_filepath']} ..."
            with Path(filepath["local_filepath"]).open("rb") as file:
                file_content_iterator = read_in_chunks(file)
                response = requests.put(url, headers=BYTES_HEADER, data=file_content_iterator)
                response.raise_for_status()
    except Exception as e:
        spinner.stop()
        raise e

    spinner.text = "Done!"
    spinner.ok("✔")


@login_required
def download(remote_path: str, local_folder: Optional[str] = None, recurse: bool = False):
    headers = auth_headers_from_local_config()

    current_dir = _get_current_dir()
    current_dir = current_dir[:-1] if current_dir.endswith("/") else current_dir
    current_dir = "/" if current_dir == "" else current_dir
    is_relative_path = not remote_path.startswith("/")
    remote_path = Path(current_dir) / Path(remote_path) if is_relative_path else Path(remote_path)
    local_folder = Path(local_folder) if local_folder else Path.cwd()

    url = f"{_BURLA_SERVICE_URL}/v1/bcs/object_info"
    response = requests.post(url, json={"remote_path": str(remote_path)}, headers=headers)
    response.raise_for_status()
    remote_object_info = response.json()
    is_folder = remote_object_info.get("type") == "folder"
    is_file = remote_object_info.get("type") == "file"

    if not local_folder.is_dir():
        print(f"local_folder `{local_folder}` is not a folder.")
        return
    elif not remote_object_info.get("exists"):
        print(f"Path `{remote_path}` not found on remote server.")
        return
    elif not (is_folder or is_file):
        raise Exception("Unexpected response from server, please email jake@burla.dev")
    elif is_folder and (not recurse):
        msg = f"Path `{remote_path}` is a directory, "
        msg += "please use `--recurse` to recursively download all files in a directory."
        print(msg)
        return
    elif is_file and recurse:
        msg = f"`--recurse` option was used but remote path `{remote_path}` is not a directory."
        print(msg)
        return
    elif is_folder and remote_object_info["contents_relative_paths"] == []:
        print(f"Remote folder `{remote_path}` is empty!")
        return

    spinner = yaspin()
    spinner.text = f"Downloading {remote_path} ..."
    spinner.start()

    try:
        if is_file:
            blob_paths_to_download = [remote_path]
        elif is_folder:
            folder_contents = remote_object_info["contents_relative_paths"]
            blob_paths_to_download = [remote_path / Path(path) for path in folder_contents]

        url = f"{_BURLA_SERVICE_URL}/v1/bcs/download_urls"
        payload = {"remote_paths": [str(path) for path in blob_paths_to_download]}
        response = requests.post(url, json=payload, headers=headers)
        response.raise_for_status()
        paths_to_urls = response.json()["paths_to_urls"]

        for remote_filepath, download_url in paths_to_urls:
            local_path = local_folder / Path(remote_filepath).relative_to(remote_path.parent)
            local_path.parent.mkdir(parents=True, exist_ok=True)

            spinner.text = f"Downloading {remote_path} to {local_path}"
            with requests.get(download_url, headers=BYTES_HEADER, stream=True) as response:
                response.raise_for_status()
                with local_path.open("wb") as file:
                    for chunk in response.iter_content(chunk_size=8192):
                        file.write(chunk)
    except Exception as e:
        spinner.stop()
        raise e

    spinner.text = "Done!"
    spinner.ok("✔")
